from pymongo import ASCENDING, DESCENDING, IndexModel

from mongodantic.model import Model, ModelNotFoundError, TModel, set_database

__all__ = [
    ASCENDING,
    DESCENDING,
    IndexModel,
    Model,
    ModelNotFoundError,
    TModel,
    set_database,
]



from typing import Optional, Sequence

from pydantic import BaseModel

from mongodantic import ASCENDING, IndexModel, Model


class User(Model):
    indexes: Sequence[IndexModel] = [
        IndexModel([("first_name", ASCENDING), ("last_name", ASCENDING)]),
        IndexModel([("external_uuid", ASCENDING)], unique=True),
        IndexModel([("last_login", ASCENDING)], expireAfterSeconds=5),
    ]

    external_uuid: str
    first_name: str
    last_name: str
    last_login: int
    new_prop: Optional[str]

    async def after_load(self):
        if not self.new_prop:
            self.new_prop = "new"


class GuestbookEntry(BaseModel):
    name: str


class Guestbook(Model):
    indexes: Sequence[IndexModel] = [
        IndexModel([("date", ASCENDING)]),
    ]

    date: str
    visitors: list[GuestbookEntry]


async def test_user_model():
    u1 = User(
        external_uuid="abc-123",
        first_name="Test",
        last_name="Testerson",
        last_login=123,
    )
    await u1.save()

    u2 = await User.get_by_id(u1.id)
    assert u2.external_uuid == u1.external_uuid
    assert u2.new_prop == "new"

    assert len(await User.find({"first_name": "Test"})) == 1

    u2.first_name = "Another"
    u2.new_prop = "old"

    await u2.save()
    await u1.reload()

    assert u1.first_name == "Another"
    assert u1.new_prop == "old"

    await u1.delete()

    assert len(await User.find({})) == 0


async def test_reload_transform():
    # In the past sub-models were not properly transformed during a .reload()
    gb = Guestbook(date="2020-01-02", visitors=[GuestbookEntry(name="Test dude")])

    await gb.save()
    await gb.reload()

    assert gb.visitors[0].__class__ == GuestbookEntry
    assert gb.visitors[0].name == "Test dude"


async def test_pagination():
    for u in range(0, 100):
        user = User(
            external_uuid=f"user-{u}",
            first_name=f"First {u + 1}",
            last_name=f"Last {u + 1}",
            last_login=u,
        )
        await user.save()

    assert await User.count() == 100
    assert await User.count({"external_uuid": "user-4"}) == 1

    users = await User.find({"last_login": {"$gte": 10}}, skip=10, limit=10)

    idx = 20
    for u in users:
        assert u.external_uuid == f"user-{idx}"
        idx += 1



import pytest
from mongomock_motor import AsyncMongoMockClient

from mongodantic.model import set_database


@pytest.fixture(autouse=True)
async def mongo_database():
    client = AsyncMongoMockClient()
    db = client["unittests"]
    set_database(db, "unittest")
    yield db
    await db["User"].delete_many({})



from abc import ABC
from typing import List, Optional, Sequence, Type, TypeVar

import motor.motor_asyncio
import pydantic
from bson import ObjectId
from bson.codec_options import CodecOptions
from pymongo import IndexModel

_DB: motor.motor_asyncio.AsyncIOMotorDatabase
_ENV: str = "unknown"
_INIT_MODELS = set()

CODEC_OPTIONS = CodecOptions(tz_aware=True)
TModel = TypeVar("TModel", bound="Model")


class ModelNotFoundError(Exception):
    pass


def set_database(db: motor.motor_asyncio.AsyncIOMotorDatabase, env: str = "unknown"):
    global _DB, _ENV
    _DB = db
    _ENV = env


async def _init_db(model: Type[TModel], collection_name):
    if _DB is None:
        raise Exception(
            "You need to call set_database before attempting to access the database"
        )

    if collection_name in _INIT_MODELS:
        return

    indexes = model.__fields__["indexes"].default

    if _ENV == "unittest":
        # mongomock doesn't support codec_options
        collection = _DB.get_collection(collection_name)
    else:
        collection = _DB.get_collection(collection_name, codec_options=CODEC_OPTIONS)

    await collection.create_indexes(indexes)

    _INIT_MODELS.add(collection_name)


class Model(pydantic.BaseModel, ABC):
    # https://pymongo.readthedocs.io/en/4.1.1/api/pymongo/operations.html#pymongo.operations.IndexModel
    indexes: Sequence[IndexModel]
    id: Optional[str]

    class Config:
        # IndexModel validation is missing for pydantic so it requires this
        arbitrary_types_allowed = True

    @classmethod
    def get_collection_name(cls) -> str:
        return cls.__name__

    @classmethod
    async def make(cls, **data) -> TModel:
        instance = cls(**data)
        await instance.after_load()
        return instance

    def get_data(self) -> dict:
        data = self.dict(by_alias=True)
        del data["id"]
        del data["indexes"]

        return data

    async def reload(self) -> None:
        coll = await self._get_collection()
        doc = await coll.find_one({"_id": ObjectId(self.id)})
        if not doc:
            raise ModelNotFoundError()

        del doc["_id"]
        item = self.__class__(id=self.id, **doc)
        for key in item.get_data():
            setattr(self, key, getattr(item, key))

        await self.after_load()

    async def after_load(self) -> None:
        pass

    async def save(self) -> None:
        data = self.get_data()
        if not self.id:
            id = ObjectId()
        else:
            id = ObjectId(self.id)

        coll = await self._get_collection()
        result = await coll.replace_one(
            {"_id": id},
            data,
            upsert=True,
        )

        if result.upserted_id:
            self.id = str(result.upserted_id)

    async def delete(self) -> bool:
        if not self.id:
            raise Exception("Can't delete a model without an ID")

        coll = await self._get_collection()
        res = await coll.delete_one({"_id": ObjectId(self.id)})
        return res.deleted_count == 1

    @classmethod
    async def find(
        cls,
        filter,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List[TModel]:
        models = []
        coll = await cls._get_collection()

        cursor = coll.find(filter)
        if skip:
            cursor.skip(skip)
        if limit:
            cursor.limit(limit)

        async for doc in cursor:
            id = str(doc["_id"])
            del doc["_id"]
            models.append(await cls.make(id=id, **doc))
        return models

    @classmethod
    async def count(cls, filter=None) -> int:
        if not filter:
            filter = {}

        coll = await cls._get_collection()
        return await coll.count_documents(filter)

    @classmethod
    async def find_one(cls, filter) -> TModel:
        coll = await cls._get_collection()
        doc = await coll.find_one(filter)
        if not doc:
            raise ModelNotFoundError()

        id = str(doc["_id"])
        del doc["_id"]

        return await cls.make(id=id, **doc)

    @classmethod
    async def get_by_id(cls, id: str) -> TModel:
        return await cls.find_one({"_id": ObjectId(id)})

    @classmethod
    async def _get_collection(cls) -> motor.motor_asyncio.AsyncIOMotorCollection:
        name = cls.get_collection_name()
        await _init_db(cls, name)

        if _ENV == "unittest":
            # mongomock doesn't support codec_options
            return _DB.get_collection(name)
        else:
            return _DB.get_collection(name, codec_options=CODEC_OPTIONS)
